/*
 Wigan Wallgate: An open-source implementation of Grand Central Dispatch
 
 Copyright (c) 2009 Matt Wright
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 */

#include "group.h"

#include <PLBlocks/Block.h>
#include <pthread.h>
#include <stdlib.h>
#include <sys/time.h>
#include <mach/mach_time.h>

struct dispatch_group_s
{
  struct dispatch_object_s object;
  int job_count;
  pthread_cond_t wait_cond; 
  dispatch_list_t suspended_blocks;
};

dispatch_group_t dispatch_group_create(void)
{
  dispatch_group_t group = (dispatch_group_t)malloc(sizeof(struct dispatch_group_s));
  memset(group, '\0', sizeof(struct dispatch_group_s));
  
  pthread_cond_init(&group->wait_cond, NULL);

  group->job_count = 0;
  group->object.retain_count = 1;
  group->suspended_blocks = dispatch_list_init();

  group->object.dealloc = Block_copy(^{
    dispatch_block_t block;
    while (block = dispatch_list_pop_from_front(group->suspended_blocks))
    {
      Block_release(block);
    }
    dispatch_release(group->suspended_blocks);
  });
  
  return group;
}

void dispatch_group_async(dispatch_group_t group, dispatch_queue_t queue, dispatch_block_t block)
{
  dispatch_retain(group);
  dispatch_group_enter(group);
  dispatch_async(queue, ^{
    block();
    dispatch_group_leave(group);
    dispatch_release(group);
  });
}

void dispatch_group_notify(dispatch_group_t group, dispatch_queue_t queue, dispatch_block_t block)
{
  // Copy the block so we can keep it around for the notify.
  dispatch_block_t heap_block = Block_copy(block);
  
  // Wrap the block up so that we can just execute this and it'll dispatch.
  dispatch_block_t notify_block = Block_copy(^{
    // Dispatch it then release it.
    dispatch_async(queue, heap_block);
    Block_release(heap_block);
  });
  
  // Pop the block on here so we can run it later.
  dispatch_list_append(group->suspended_blocks, notify_block);
}

long dispatch_group_wait(dispatch_group_t group, dispatch_time_t timeout)
{
  pthread_mutex_t mutex;
  pthread_mutex_init(&mutex, NULL);

  // the timeout is in nanoseconds but timed wait works in seconds
  struct timeval tv;
  struct timespec to;
  gettimeofday(&tv, NULL);
  
  // the timeout is in mach_time, so we need to convert it to a relative
  // number of nanoseconds to wait
  if (timeout == DISPATCH_TIME_NOW)
  {
    timeout = dispatch_time(timeout, 0);
  }

  dispatch_time_t towait = timeout - dispatch_time(DISPATCH_TIME_NOW, 0);
  to.tv_nsec = ((tv.tv_usec * NSEC_PER_USEC) + (towait % NSEC_PER_SEC)) % NSEC_PER_SEC;
  to.tv_sec = tv.tv_sec + (towait / NSEC_PER_SEC) + (((tv.tv_usec * NSEC_PER_USEC) + (towait % NSEC_PER_SEC)) / NSEC_PER_SEC);
  
  int res;
  if (timeout == DISPATCH_TIME_FOREVER)
  {
    res = pthread_cond_wait(&group->wait_cond, &mutex);
  }
  else
  {
    res = pthread_cond_timedwait(&group->wait_cond, &mutex, &to);
  }
  
  if (res == EINVAL)
  {
    WW_ERROR("pthread_cond_timedwait returned EINVAL\n");
    return -1;
  }
  else if (res == ETIMEDOUT)
  {
    return -1;
  }
  return 0;
}

void dispatch_group_enter(dispatch_group_t group)
{
  // increment the job counter
  OSAtomicIncrement32(&group->job_count);
}

void dispatch_group_leave(dispatch_group_t group)
{
  int count = OSAtomicDecrement32(&group->job_count);
  if (count == 0)
  {
    // Pull the blocks off the list waiting for notify
    dispatch_block_t block;
    while (block = dispatch_list_pop_from_front(group->suspended_blocks))
    {
      // These blocks are a wrapped dispatch_async, so fire it off. 
      block();
    }
    
    // Signal anything we've got waiting on a _wait.
    pthread_cond_broadcast(&group->wait_cond);
  }
}
