/*
 Wigan Wallgate: An open-source implementation of Grand Central Dispatch
 
 Copyright (c) 2009 Matt Wright
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 */

#include <stdlib.h>
#include <PLBlocks/Block.h>

#include "list.h"

dispatch_list_t dispatch_list_init()
{
  dispatch_list_t list = CFAllocatorAllocate(NULL, sizeof(struct dispatch_list_s), 0);
  memset(list, '\0', sizeof(struct dispatch_list_s));
  
  CFArrayCallBacks callbacks;
  
  // clear out the callbacks, we're not interested atm
  memset(&callbacks, '\0', sizeof(CFArrayCallBacks));
  
  // create the array
  list->array = CFArrayCreateMutable(NULL, 0, &callbacks);
  
  // init the spinlock
  list->lock = OS_SPINLOCK_INIT;
  
  list->object.retain_count = 1;
  list->object.dealloc = Block_copy(^{
    CFRelease(list->array);
  });
  
  return list;
}

unsigned int dispatch_list_count(dispatch_list_t list)
{
  // gah, spin lock even this :(
  OSSpinLockLock(&list->lock);
  
  unsigned int count = CFArrayGetCount(list->array);
  OSSpinLockUnlock(&list->lock);
  
  return count;
}

void dispatch_list_insert(dispatch_list_t list, unsigned int index, void *data)
{
  // lock the spinlock
  OSSpinLockLock(&list->lock);
  
  // do i really have to explain this
  CFArrayInsertValueAtIndex(list->array, index, data);
  
  OSSpinLockUnlock(&list->lock);
}

void dispatch_list_insert_sorted(dispatch_list_t list, void *data, int (^block)(void* a, void* b))
{
  OSSpinLockLock(&list->lock);
  
  for (int i=0; i<CFArrayGetCount(list->array); i++)
  {
    void *a = (void*)CFArrayGetValueAtIndex(list->array, i);
    void *b = data;
    
    if (block(a, b) > -1)
    {
      CFArrayInsertValueAtIndex(list->array, i, data);
      OSSpinLockUnlock(&list->lock);
      return;
    }
  }
  
  CFArrayAppendValue(list->array, data);
  
  OSSpinLockUnlock(&list->lock);
}

void dispatch_list_append(dispatch_list_t list, void *data)
{
  OSSpinLockLock(&list->lock);
  CFArrayAppendValue(list->array, data);
  OSSpinLockUnlock(&list->lock);
}

void dispatch_list_remove_at_index(dispatch_list_t list, unsigned int index)
{
  OSSpinLockLock(&list->lock);
  if (index < CFArrayGetCount(list->array))
  {
    CFArrayRemoveValueAtIndex(list->array, index);
  }
  OSSpinLockUnlock(&list->lock);
}

void dispatch_list_remove_value(dispatch_list_t list, void *data)
{
  OSSpinLockLock(&list->lock);
  
  for (int i=0; i<CFArrayGetCount(list->array); i++)
  {
    if (CFArrayGetValueAtIndex(list->array, i) == data)
    {
      CFArrayRemoveValueAtIndex(list->array, i);
      OSSpinLockUnlock(&list->lock);
      return;
    }
  }
  
  OSSpinLockUnlock(&list->lock);
}

void* dispatch_list_value_at_index(dispatch_list_t list, unsigned int index)
{
  OSSpinLockLock(&list->lock);
  void* ret = NULL;

  if (CFArrayGetCount(list->array))
  {
    // use -1 for last ;)
    if (index == -1)
    {
      ret = (void*)CFArrayGetValueAtIndex(list->array, CFArrayGetCount(list->array) - 1);
    }
    ret = (void*)CFArrayGetValueAtIndex(list->array, index);
  }
  
  OSSpinLockUnlock(&list->lock);
  return ret;
}

void* dispatch_list_pop_from_front(dispatch_list_t list)
{
  OSSpinLockLock(&list->lock);
  
  void *item = NULL;
  if (CFArrayGetCount(list->array) > 0)
  {
    item = (void*)CFArrayGetValueAtIndex(list->array, 0);
    CFArrayRemoveValueAtIndex(list->array, 0);
  }
  OSSpinLockUnlock(&list->lock);
  
  return item;
}
