/*
 Wigan Wallgate: An open-source implementation of Grand Central Dispatch
 
 Copyright (c) 2009 Matt Wright
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 */

#ifndef __DISPATCH_MANAGER_H__
#define __DISPATCH_MANAGER_H__

#include <stdio.h>
#include <pthread.h>

#include <WiganWallgate/base.h>
#include <WiganWallgate/queue.h>

#include <mach/machine.h>
#include <mach/processor_info.h>

/*
 This is the struct used to control a worker thread. *amend desc later*
 */
struct dispatch_worker_thread
{
  // the actual thread
  pthread_t thread;
};

/*
 This is the "manager". Eventually, all our queues will end up here and we'll keep track of the
 pthreads we spawn to create our thread pool.
 */
struct dispatch_manager
{
  // keep track of these
  unsigned int ncpus;
  unsigned int nworkers;
  
  struct dispatch_worker_thread *workers;
  
  // dealing with demand threads, ones to pick up the slack when you've spare cpu
  unsigned int nworking;
  unsigned int ndworkers;
  unsigned int ndemand;
  
  // a running average of cpu load
  processor_cpu_load_info_t processorLoadInfo;
  mach_msg_type_number_t processorMessageCount;
  
  // a condvar for all the workers to wait on for new jobs
  pthread_cond_t cond;
  
  // workers hold this mutex while they look for jobs to do
  pthread_mutex_t mutex;
  
  // high priority queue
  dispatch_queue_t high_queue;
  
  // the default priority concurrent queue
  dispatch_queue_t default_queue;
  
  // low priority queue
  dispatch_queue_t low_queue;
  
  // a special queue for the main thread that gets ignored by
  // the serial queue dispatch code
  dispatch_queue_t main_thread_queue;
};

/*
 This gets the global (ick) manager. Everyone talks to him.
 */
struct dispatch_manager* dispatch_manager_get_global_manager();

/*
 This one is not in GCD in Snow Leopard. We don't have the advantage of being able to instantly
 tie ourselves into libSystem/Cocoa and magic our way into runloops etc. So, we're gonna mark them
 as constructors in gcc and they'll get run before main().
 */
void dispatch_manager_init() __attribute__((constructor));

/*
 We can't avoid having to make the use setup the main thread consumer because
 we can't hook into Foundation ourselves and make it automagically appear like
 in Grand Central. Boo.
 */
void dispatch_manager_runloop_init() __attribute__((constructor));

/*
 This function lets us manage the size of the thread pool depending on the load on the CPU, if we've
 got free blocks to run in the concurrent queues and we've got spare CPU power then we start to turn
 up the number of threads
 */
void dispatch_manager_thread_monitor();

/*
 Starts a given worker thread
*/
void dispatch_worker_thread_start(struct dispatch_worker_thread *worker);

/*
 The processing loop for each worker thread
*/
void* dispatch_worker_thread_runloop(void *arg);

/*
 Notify the manager that there is work to do
*/
void dispatch_manager_notify(struct dispatch_manager *manager, dispatch_queue_t queue);

/*
 Runs the queue in the current thread till it is empty
*/
void dispatch_manager_drain_queue(dispatch_queue_t queue);

#endif // __DISPATCH_MANAGER_H__
