/*
 Wigan Wallgate: An open-source implementation of Grand Central Dispatch
 
 Copyright (c) 2009 Matt Wright
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 */

#include "manager.h"

#ifdef TARGET_OS_IPHONE
#include <Foundation/Foundation.h>
#endif

#include <PLBlocks/Block.h>
#include <sys/param.h>
#include <sys/sysctl.h>
#include <stdlib.h>
#include <mach/mach.h>

/*
 Nasty nasty hacks, I dislike doing this but it is the easiest way atm.
 */
struct dispatch_manager* dispatch_manager_get_global_manager()
{
  static struct dispatch_manager global_manager;
  return &global_manager;  
}

void dispatch_manager_init()
{
  WW_DBG("global manager init\n");

  // get the manager, we can assume it isn't initialised yet.
  struct dispatch_manager *manager = dispatch_manager_get_global_manager();
  
  // first get the number of cpu's on the system.
  size_t ncpus_size = sizeof(manager->ncpus);
  
  if (sysctlbyname("hw.ncpu", &manager->ncpus, &ncpus_size, NULL, 0))
  {
    // couldn't get the cpu count so stick with one
    manager->ncpus = 1;
  }
  
  // for now, go with the naive CPUS+2 approach
  manager->nworkers = manager->ncpus + 2;
  manager->nworking = 0;
  manager->ndemand = 0;
  manager->ndworkers = 0;
  
  // init the initial processing tick counts
  natural_t processorCount;
  kern_return_t err = host_processor_info(mach_host_self(), PROCESSOR_CPU_LOAD_INFO, &processorCount, (processor_info_array_t*)&manager->processorLoadInfo, &manager->processorMessageCount);
  if (err != KERN_SUCCESS)
  {
    WW_ERROR("failed to read out the processor load counts, non-fatal.\n");
  }
    
  // before we fire up the workers, setup the cond+mutex
  if (pthread_cond_init(&manager->cond, NULL))
  {
    WW_ERROR("failed to init manager pthread_cont_t\n");
    return;
  }
  
  if (pthread_mutex_init(&manager->mutex, NULL))
  {
    WW_ERROR("failed to init manager pthread_mutex_t\n");
    return;
  }
  
  // allocate and initialise the default queues. the values for the queue_t's are arranged such that
  // a completely zeroed queue is valid as a the initial state of a concurrent queue.
  manager->default_queue = dispatch_queue_create_concurrent("uk.co.sysctl.wiganwallgate.default_queue", 0);
  manager->high_queue = dispatch_queue_create_concurrent("uk.co.sysctl.wiganwallgate.high_queue", 0);
  manager->low_queue = dispatch_queue_create_concurrent("uk.co.sysctl.wiganwallgate.low_queue", 0);
  manager->main_thread_queue = dispatch_queue_create("uk.co.sysctl.wiganwallgate.main_thread_queue", 0);
  
  WW_DBG("found %d cpu's, initialising %d worker threads\n", manager->ncpus, manager->nworkers);  
  
  // allocate the workers
  manager->workers = (struct dispatch_worker_thread*)malloc(sizeof(struct dispatch_worker_thread) * manager->nworkers);
  
  // start each of them
  for (int i = 0; i < manager->nworkers; i++)
  {
    dispatch_worker_thread_start(&manager->workers[i]);
  }
}

// just to pass to the runloop observer below
void RunloopCallback(CFRunLoopObserverRef ref, CFRunLoopActivity activity, void *context)
{
  dispatch_manager_drain_queue(dispatch_manager_get_global_manager()->main_thread_queue);
}

void RunloopTimerCallback(CFRunLoopTimerRef timer, void *info)
{
  dispatch_manager_drain_queue(dispatch_manager_get_global_manager()->main_thread_queue);
}

void dispatch_manager_runloop_init()
{
  CFRunLoopObserverRef obRef = CFRunLoopObserverCreate(NULL,
                                                       kCFRunLoopBeforeWaiting,
                                                       true,
                                                       0,
                                                       RunloopCallback, 
                                                       NULL);
  CFRunLoopAddObserver(CFRunLoopGetMain(), obRef, kCFRunLoopCommonModes);
}

void dispatch_manager_notify(struct dispatch_manager *manager, dispatch_queue_t queue)
{
  // so, check if the queue we've been asked to poke from is one of our default queues
  if ((queue != manager->default_queue) && (queue != manager->low_queue) && (queue != manager->high_queue) && (queue != manager->main_thread_queue))
  {
    // once we know it isn't, then it'll be someone else's queue.
    if (queue->busy == DISPATCH_QUEUE_IDLE)
    {      
      // get the first block in the queue
      if (dispatch_list_count(queue->list) > 0)
      {
        // if the queue is idle, we need to move a block off it into the concurrent
        // queue.
        if (queue->type == DISPATCH_QUEUE_TYPE_SERIAL)
        {
          if (!OSAtomicCompareAndSwap32Barrier(DISPATCH_QUEUE_IDLE, DISPATCH_QUEUE_BUSY, (int32_t*)&queue->busy))
          {
            // we couldn't do the compare exchange. its changed and we're busy now, don't dispatch
            // just do the pthread_cond and get out of here
            pthread_cond_signal(&manager->cond);
            return;
          }
        }
        
        dispatch_block_t block = dispatch_list_pop_from_front(queue->list);
      
        // we're in trouble if the list didn't have any work to do but check anyway
        if (block != NULL)
        {
          // we're gonna reference the queue from teh block so retain it here
          dispatch_retain(queue);
        
          dispatch_block_t dispatch_block = Block_copy(^{
            // run the block
            block();
            
            // release the block we just run
            Block_release(block);
            
            // now we need to poke the queue we just finished executing from
            queue->busy = DISPATCH_QUEUE_IDLE;
            
            // do a manager notify to say we might have more work to do
            dispatch_manager_notify(manager, queue);
            
            // don't need it anymore
            dispatch_release(queue);
          });
          
          // now put that block on a concurrent queue to execute
          dispatch_list_append(manager->default_queue->list, dispatch_block);
          
        } // if (block != NULL)
        else
        {
          // oof fuck. We got the lock for a serial queue above and now we can't use it. If it was a concurrent
          // queue this will just do nothign anyway.
          OSAtomicCompareAndSwap32Barrier(DISPATCH_QUEUE_BUSY, DISPATCH_QUEUE_IDLE, (int32_t*)&queue->busy);
          WW_ERROR("arggh");
        }
        
      } // if (queue->list)
            
    } // if (queue->busy == DISPATCH_QUEUE_IDLE)
  }
  
  if (pthread_cond_signal(&manager->cond))
  {
    WW_ERROR("failed to signal manager of new jobs\n");
  }
  
  // if this was actually a main thread job, we'll wake him up too. The signal above is pretty harmless anyway.
  if (queue == manager->main_thread_queue)
  {
    CFRunLoopWakeUp(CFRunLoopGetMain());
  }
  
}

dispatch_block_t _dispatch_worker_dequeue_block(struct dispatch_manager *manager)
{
  dispatch_block_t block = NULL;
  
  // first try, see if we've got something on the high queue.
  if (dispatch_list_count(manager->high_queue->list) > 0)
  {
    block = dispatch_list_pop_from_front(manager->high_queue->list);
  }
  
  // if not, try the default queue
  if ((block == NULL) && (dispatch_list_count(manager->default_queue->list) > 0))
  {
    block = dispatch_list_pop_from_front(manager->default_queue->list);
  }
  
  // if not, try the low queue
  if ((block == NULL) && (dispatch_list_count(manager->low_queue->list) > 0))
  {
    block = dispatch_list_pop_from_front(manager->low_queue->list);
  }
  
  return block;
}

void dispatch_worker_thread_start(struct dispatch_worker_thread *worker)
{  
  // all we do here at the moment is spawn a new pthread and have rid of it
  if (pthread_create(&worker->thread, NULL, dispatch_worker_thread_runloop, worker))
  {
    WW_ERROR("failed to initalise worker thread.\n");
  }
}

void* dispatch_worker_thread_runloop(void *arg)
{
  //struct dispatch_worker_thread *worker = arg;
  struct dispatch_manager *manager = dispatch_manager_get_global_manager();
    
  while (1)
  {
    // do stuff to find a job to do
    dispatch_block_t block = _dispatch_worker_dequeue_block(manager);
    
    pthread_mutex_unlock(&manager->mutex);
    
    if (block != NULL)
    {
      // mark us as working
      OSAtomicIncrement32((int32_t*)&manager->nworking);
      
      // We need one of these really
      NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
      
      // we've got work, woot
      block();
      
      // now we don't
      [pool release];
      
      // this block should be heap allocated, so release it here
      Block_release(block);
      
      // and now as not working
      OSAtomicDecrement32((int32_t*)&manager->nworking);
    }
    else
    {
      // we got here, so we've no work to do.

      if (pthread_cond_wait(&manager->cond, &manager->mutex))
      {
        WW_ERROR("uh oh, pthread_cond_wait failed in worker runloop\n");
      }
    }
  }
  
  WW_DBG("runloop exited\n");
  
  return NULL;
}

void dispatch_manager_drain_queue(dispatch_queue_t queue)
{
  while (dispatch_list_count(queue->list) > 0)
  {
    dispatch_block_t block = dispatch_list_pop_from_front(queue->list);
    if (block != NULL)
    {
      block();
      Block_release(block);
    }
  }
}

void* dispatch_demand_thread_runloop(void *arg)
{
  struct dispatch_manager *manager = dispatch_manager_get_global_manager();
  
  // make sure the manager knows we're here
  OSAtomicIncrement32((int32_t*)&manager->ndworkers);
  
  // we come up while running, so we should take the lock just in case
  pthread_mutex_lock(&manager->mutex);
  
  while (1)
  {
    dispatch_block_t block = _dispatch_worker_dequeue_block(manager);
    
    pthread_mutex_unlock(&manager->mutex);
    
    if (block != NULL)
    {
      // mark us as working
      OSAtomicIncrement32((int32_t*)&manager->nworking);
      
      // We need one of these really
      NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
      
      // we've got work, woot
      block();
      
      // now we don't
      [pool release];
      
      // this block should be heap allocated, so release it here
      Block_release(block);
      
      // and now as not working
      OSAtomicDecrement32((int32_t*)&manager->nworking);
    }
    else
    {
      // before we wait on the condvar, we should check the number of demand workers isn't supposed to have
      // been reduced.
      int demand = manager->ndemand, workers = manager->ndworkers;
      
      if (demand < workers)
      {
        if (OSAtomicCompareAndSwap32Barrier(workers, workers-1, (int32_t*)&manager->ndworkers))
        {
          // if this succeeded, then we just decided to cull ourselves
          pthread_exit(NULL);
        }
      }
      
      if (pthread_cond_wait(&manager->cond, &manager->mutex))
      {
        WW_ERROR("uh oh, pthread_cond_wait failed in demand runloop\n");
      }
    }
    
  }
  
}

void dispatch_manager_thread_monitor()
{
  struct dispatch_manager *manager = dispatch_manager_get_global_manager();
  
  // get the cpu load
  natural_t processorCount;
  processor_cpu_load_info_t processorLoadInfo;
  mach_msg_type_number_t processorMessageCount;
  
  kern_return_t err = host_processor_info(mach_host_self(), PROCESSOR_CPU_LOAD_INFO, &processorCount, (processor_info_array_t*)&processorLoadInfo, &processorMessageCount);
  if (err != KERN_SUCCESS)
  {
    // bomb out, try again next loop
    return;
  }
  
  float load = 0.0;
  
  for (natural_t i = 0; i < processorCount; i++)
  {
    uint64_t user = 0, system = 0, idle = 0, total = 0;
    if (processorLoadInfo[i].cpu_ticks[CPU_STATE_SYSTEM] >= manager->processorLoadInfo[i].cpu_ticks[CPU_STATE_SYSTEM])
    {
      system = processorLoadInfo[i].cpu_ticks[CPU_STATE_SYSTEM] - manager->processorLoadInfo[i].cpu_ticks[CPU_STATE_SYSTEM];
    }
    else
    {
      system = processorLoadInfo[i].cpu_ticks[CPU_STATE_SYSTEM] + (UINT_MAX - manager->processorLoadInfo[i].cpu_ticks[CPU_STATE_SYSTEM] + 1);
    }
    
    if (processorLoadInfo[i].cpu_ticks[CPU_STATE_USER] >= manager->processorLoadInfo[i].cpu_ticks[CPU_STATE_USER])
    {
      user = processorLoadInfo[i].cpu_ticks[CPU_STATE_USER] - manager->processorLoadInfo[i].cpu_ticks[CPU_STATE_USER];
    }
    else
    {
      user = processorLoadInfo[i].cpu_ticks[CPU_STATE_USER] + (UINT_MAX - manager->processorLoadInfo[i].cpu_ticks[CPU_STATE_USER] + 1);
    }

    if (processorLoadInfo[i].cpu_ticks[CPU_STATE_IDLE] >= manager->processorLoadInfo[i].cpu_ticks[CPU_STATE_IDLE])
    {
      idle = processorLoadInfo[i].cpu_ticks[CPU_STATE_IDLE] - manager->processorLoadInfo[i].cpu_ticks[CPU_STATE_IDLE];
    }
    else
    {
      idle = processorLoadInfo[i].cpu_ticks[CPU_STATE_IDLE] + (UINT_MAX - manager->processorLoadInfo[i].cpu_ticks[CPU_STATE_IDLE] + 1);
    }    
    
    total = system + user + idle;
    load = (float)(system + user) / (float)total;
  }
  
  // now we've got a load average. decide if we need to increase the number of threads.
  int pending_jobs = dispatch_list_count(manager->high_queue->list) + dispatch_list_count(manager->default_queue->list) + dispatch_list_count(manager->low_queue->list);

  if (manager->nworking == manager->nworkers + manager->ndworkers)
  {
    if ((pending_jobs > 0) && (load < 0.75))
    {
      OSAtomicIncrement32((int32_t*)&manager->ndemand);
      
      pthread_t thread;
      pthread_create(&thread, NULL, dispatch_demand_thread_runloop, NULL);
    }
    
  }
  else if ((manager->nworking < manager->nworkers + manager->ndworkers) && (manager->ndemand > 0))
  {
    if (pending_jobs == 0)
    {
      OSAtomicDecrement32((int32_t*)&manager->ndemand);
      
      // wake up all the threads and see who dies. This actually wakes the non-demand workers too
      // but they'll just go straight back to sleep again. Or service another block.
      pthread_cond_broadcast(&manager->cond);
    }
    else
    {
      WW_DBG("pending jobs > 0 but nworking != nworkers.\n");
    }
    
  }
  
  for (natural_t i = 0; i < processorCount; i++)
  {
    for (natural_t j = 0; j < CPU_STATE_MAX; j++)
    {
      manager->processorLoadInfo[i].cpu_ticks[j] = processorLoadInfo[i].cpu_ticks[j];
    }
  }
  
  // dealloc the memory from processor info
  vm_deallocate(mach_task_self(), (vm_address_t)processorLoadInfo, (vm_size_t)(processorMessageCount * sizeof(*processorLoadInfo)));
}
