/*
 Wigan Wallgate: An open-source implementation of Grand Central Dispatch
 
 Copyright (c) 2009 Matt Wright
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 */

#include <WiganWallgate/object.h>
#include <PLBlocks/Block.h>
#include <libkern/OSAtomic.h>
#include <stdlib.h>

typedef struct dispatch_object_s* dispatch_object_tt;

void dispatch_retain(dispatch_object_tt object)
{
  OSAtomicIncrement32(&object->retain_count);
}

void dispatch_release(dispatch_object_tt object)
{
  int count = OSAtomicDecrement32(&object->retain_count);
  if (count == 0)
  {
    dispatch_function_t finaliser = object->finaliser;
    void *context = object->context;
    
    // do internal dealloc if we've got one
    if (object->dealloc)
    {
      object->dealloc();
      Block_release(object->dealloc);
    }
    
    if (object->suspend)
    {
      object->suspend();
      Block_release(object->suspend);
    }
    
    if (object->resume)
    {
      Block_release(object->resume);
    }
    
    // do the actual release
    free(object);
    
    // call the finaliser function
    if (finaliser != NULL)
    {
      finaliser(context);
    }
  }
}

void dispatch_suspend(dispatch_object_tt object)
{
  if (object->suspend)
  {
    int count = object->suspend_count;
    
    OSAtomicIncrement32(&object->suspend_count);
    
    // we weren't suspended before
    if (count == 0)
    {
      object->suspend();
    }
  }
  else
  {
    WW_ERROR("dispatch_suspect: this object does not implement suspend\n");
  }
}

void dispatch_resume(dispatch_object_tt object)
{
  if (object->resume)
  {
    int count = OSAtomicDecrement32(&object->suspend_count);
    
    // we were suspended before
    if (count == 0)
    {
      object->resume();
    }
  }
  else
  {
    WW_ERROR("dispatch_suspect: this object does not implement suspend\n");
  }
}
