/*
 Wigan Wallgate: An open-source implementation of Grand Central Dispatch
 
 Copyright (c) 2009 Matt Wright
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
*/

#include <PLBlocks/Block.h>
#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <mach/mach_time.h>

#include "queue.h"
#include "manager.h"
#include "sources.h"

dispatch_queue_t dispatch_get_main_queue()
{
  return dispatch_manager_get_global_manager()->main_thread_queue;
}

dispatch_queue_t dispatch_get_global_queue(long priority, unsigned long flags)
{
  // needs more work, just returns the default queue for any priority level atm
  switch (priority)
  {
    case DISPATCH_QUEUE_PRIORITY_DEFAULT:
      return dispatch_manager_get_global_manager()->default_queue;
    case DISPATCH_QUEUE_PRIORITY_HIGH:
      return dispatch_manager_get_global_manager()->high_queue;
    case DISPATCH_QUEUE_PRIORITY_LOW:
      return dispatch_manager_get_global_manager()->low_queue;
  }
  return dispatch_manager_get_global_manager()->default_queue;
}

dispatch_queue_t dispatch_queue_create_concurrent(const char *label, dispatch_queue_attr_t attr)
{
  dispatch_queue_t queue = (dispatch_queue_t)malloc(sizeof(struct dispatch_queue_s));
  memset(queue, '\0', sizeof(struct dispatch_queue_s));
  
  queue->object.retain_count = 1; // these probably never actually get released
  queue->list = dispatch_list_init();
  
  if (label)
  {
    queue->name = strdup(label);
  }
  
  return queue;
}

dispatch_queue_t dispatch_queue_create(const char *label, dispatch_queue_attr_t attr)
{
  dispatch_queue_t queue = (dispatch_queue_t)malloc(sizeof(struct dispatch_queue_s));
  memset(queue, '\0', sizeof(struct dispatch_queue_s));
  
  queue->name = NULL;
  if (label)
  {
    queue->name = strdup(label);
  }
  
  queue->object.retain_count = 1;
  queue->list = dispatch_list_init();
  
  queue->type = DISPATCH_QUEUE_TYPE_SERIAL;
  queue->busy = DISPATCH_QUEUE_IDLE;
  
  queue->object.dealloc = Block_copy(^{
    if (queue->name)
    {
      free(queue->name);
    }
    dispatch_release(queue->list);
  });
  
  return queue;
}

void dispatch_async(dispatch_queue_t queue, dispatch_block_t block)
{
  // first thing, thanks http://thirdcog.eu/pwcblocks/
  dispatch_block_t heap_block = Block_copy(block);
  
  // retain the queue, so it can't disappear under us till the block has run
  dispatch_retain(queue);
  
  // right, unlike sync we need to do all the clean up in the block we put 
  // on the queue.
  
  dispatch_block_t async_block = Block_copy(^{
    // run the block
    heap_block();
    
    // now clean up
    Block_release(heap_block);
    
    // release the queue too
    dispatch_release(queue);
  });
  
  // whack it on the list
  dispatch_list_append(queue->list, async_block);
  
  // tell them we've got work
  dispatch_manager_notify(dispatch_manager_get_global_manager(), queue);
}

void dispatch_async_f(dispatch_queue_t queue, void *context, dispatch_function_t work)
{
  dispatch_async(queue, ^{
    work(context);
  });
}

void dispatch_sync(dispatch_queue_t queue, dispatch_block_t block)
{
  // we don't actually need to block copy the block here, its a waste of resources
  // as this function won't go out of scope until we've finished executing the
  // block. Neither will the calling function.
  
  // retain the queue, so it can't disappear under us till the block has run
  dispatch_retain(queue);
  
  // we're gonna put it in the queue but not in the way you think.
  
  // first, get a condvar and a mutex (actually we don't give a crap about
  // the mutex but meh).
  pthread_cond_t cond, *pCond = &cond;
  pthread_mutex_t mutex, *pMutex = &mutex;
  
  pthread_cond_init(&cond, NULL);
  pthread_mutex_init(&mutex, NULL);
  
  // lock it, this'll stop the block completing till we're ready to condwait
  pthread_mutex_lock(&mutex);

  // doubly annoying, we can't use a stack block here because it has to be
  // released by the normal runloop code. Which also means you can't use __block
  // qualifiers on the cond/mutex above. So pass in pointers to them instead.
  dispatch_block_t sync_block = Block_copy(^{
    // lock the block
    pthread_mutex_lock(pMutex);
    
    // firstly execute the block
    block();
    
    // done
    dispatch_release(queue);

    // now signal the sync that we're done
    pthread_mutex_unlock(pMutex);
    pthread_cond_signal(pCond);
  });
  
  dispatch_list_append(queue->list, sync_block);
  
  // kick the thread manager to say we've got new work to do
  dispatch_manager_notify(dispatch_manager_get_global_manager(), queue);
  
  // now, we've put it on the queue. we wait.
  pthread_cond_wait(&cond, &mutex);
}

void dispatch_sync_f(dispatch_queue_t queue, void *context, dispatch_function_t work)
{
  dispatch_sync(queue, ^{
    work(context);
  });
}

void dispatch_apply(size_t iterations, dispatch_queue_t queue, void (^block)(size_t))
{
  // I could implement this using groups but the help docs say it is in queue.h, so I'm gonna
  // do it without
  size_t i, __block left = iterations;
  
  pthread_mutex_t mutex;
  pthread_cond_t __block cond;
  
  pthread_cond_init(&cond, NULL);
  pthread_mutex_init(&mutex, NULL);
  
  for (i=0; i<iterations; i++)
  {
    dispatch_async(queue, ^{
      block(i);
      
      // decrement and signal if done
      if (OSAtomicDecrement32((int32_t*)&left) == 0)
      {
        pthread_cond_signal(&cond);
      }
    });
  }
  
  if (left > 0)
  {
    pthread_cond_wait(&cond, &mutex);
  }
}

void dispatch_apply_f(size_t iterations, dispatch_queue_t queue, void *context, void (*work)(void *, size_t))
{
  dispatch_apply(iterations, queue, ^(size_t i){
    work(context, i);
  });
}

void dispatch_after(dispatch_time_t when, dispatch_queue_t queue, dispatch_block_t block)
{
  // this method is better, it'll be even better when timer sources don't take a thread
  // each
  dispatch_source_t source = dispatch_source_create(DISPATCH_SOURCE_TYPE_TIMER, 0, 0, queue);
  
  // copy the block
  dispatch_block_t heap_block = Block_copy(block);
  
  dispatch_source_set_event_handler(source, ^{
    heap_block();
    
    Block_release(heap_block);
    
    dispatch_source_cancel(source);
  });
  
  dispatch_source_set_cancel_handler(source, ^{
    dispatch_release(source);
  });
  
  dispatch_source_set_timer(source, when, 0, 0);
  dispatch_resume(source);
}

void dispatch_after_f(dispatch_time_t when, dispatch_queue_t queue, void *context, dispatch_function_t work)
{
  dispatch_after(when, queue, ^{
    work(context);
  });
}