/*
 Wigan Wallgate: An open-source implementation of Grand Central Dispatch
 
 Copyright (c) 2009 Matt Wright
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 */

#include <PLBlocks/Block.h>
#include <WiganWallgate/list.h>
#include <pthread.h>
#include <sys/select.h>
#include <sys/event.h>
#include <mach/mach_time.h>

#include <WiganWallgate/sources.h>
#include <WiganWallgate/queue.h>
#include <WiganWallgate/manager.h>

#pragma mark dispatch source type

struct dispatch_source_type_s
{
  void (^registerBlock)(dispatch_source_t source, uintptr_t handle, unsigned long mask);
  void (^unregisterBlock)(dispatch_source_t source);
  void (^dispatchEvent)(dispatch_source_t source, struct kevent event);
};

#pragma mark dispatch source

struct dispatch_source_s
{
  struct dispatch_object_s object;
  
  // cancelled
  int cancelled;
  int running;
  pthread_cond_t cancel_cond;
  
  // hold some useful info
  uintptr_t handle;
  unsigned long mask;
  dispatch_queue_t queue;
  unsigned long data;
  
  // random context
  void *context;
  
  // timer specific stuff
  dispatch_time_t start;
  uint64_t interval;
  uint64_t leeway;
  
  // and the type of source
  dispatch_source_type_t type;
  
  // and the handler
  dispatch_block_t event_handler;
  dispatch_block_t cancel_handler;
};

#pragma mark kevent source

struct dispatch_kevent_source_s
{
  int kq;
  pthread_t thread;
  dispatch_list_t timers;
};

typedef struct dispatch_kevent_source_s* dispatch_kevent_source_t;

dispatch_kevent_source_t dispatch_source_get_global_kevent_source()
{
  static struct dispatch_kevent_source_s source;
  return &source;
}

void* dispatch_source_kevent_runloop(void *context)
{
  dispatch_kevent_source_t source = context;
  
  while (1)
  {
    struct kevent events[1];
    struct timespec ts = { 5, 0 };
    
    int nfired = kevent(source->kq, NULL, 0, events, sizeof(events) / sizeof(events[0]), &ts);
    if (nfired > 0)
    {
      dispatch_source_t event_source = events[0].udata;
      event_source->type->dispatchEvent(event_source, events[0]);
    }
    else {
      // call the thread monitor to look if we need more threads to do the work.
      dispatch_manager_thread_monitor();
    }
  }
  
  return NULL;
}

void dispatch_source_kevent_init() __attribute__((constructor));
void dispatch_source_kevent_init()
{
  dispatch_kevent_source_t source = dispatch_source_get_global_kevent_source();
  source->kq = kqueue();
  source->timers = dispatch_list_init();
  
  if (pthread_create(&source->thread, NULL, dispatch_source_kevent_runloop, source))
  {
    WW_ERROR("kevent source runloop failed to start\n");
  }
}

void _dispatch_source_async(dispatch_source_t source, dispatch_block_t block)
{
  // grr we could possibly race here because the block we run reschedules us, so because
  // of that, I'm gonna check we're not cancelled before we run the block. This means
  // you'll lose the last event if you cancelled when its just fired.
  
  dispatch_block_t heap_block = Block_copy(block);
  source->running = 1;
  
  dispatch_async(source->queue, ^{
    if (source->cancelled != 0)
    {
      // we were cancelled, signal that we're done
      pthread_cond_signal(&source->cancel_cond);
    }
    else
    {
      heap_block();
    }
    Block_release(heap_block);

    source->running = 0;
  });
}

#pragma mark predefined sources

// ---
// predefined sources

#pragma mark read source

const struct dispatch_source_type_s _dispatch_source_type_read = {
  
  // I'm kinda channeling the linux source code about now ;)
  .registerBlock = ^(dispatch_source_t source, uintptr_t handle, unsigned long mask){
    struct kevent event;
    EV_SET(&event, handle, EVFILT_READ, EV_ADD | EV_RECEIPT | EV_ONESHOT, 0, 0, source);
    
    dispatch_kevent_source_t kevent_source = dispatch_source_get_global_kevent_source();
    int res = kevent(kevent_source->kq, &event, 1, &event, 1, NULL);
    if (res < 0)
    {
      WW_ERROR("adding kevent for read source failed: %d, %s\n", errno, strerror(errno));
    }
  },
  
  .unregisterBlock = ^(dispatch_source_t source){
    struct kevent event;
    EV_SET(&event, source->handle, EVFILT_READ, EV_DELETE, 0, 0, NULL);
    
    dispatch_kevent_source_t kevent_source = dispatch_source_get_global_kevent_source();
    int res = kevent(kevent_source->kq, &event, 1, &event, 1, NULL);
    if (res < 0)
    {
      WW_ERROR("unregistering kevent failed: %d, %s\n", errno, strerror(errno));
    }
  },
  
  .dispatchEvent = ^(dispatch_source_t source, struct kevent event){
    dispatch_retain(source);
    source->data = event.data;
    
    _dispatch_source_async(source, ^{
      source->event_handler();
      
      if (!source->cancelled && (source->object.suspend_count == 0))
      {
        // once we've handled, we need to put the event back into the event queue
        struct kevent event;
        EV_SET(&event, source->handle, EVFILT_READ, EV_ADD | EV_RECEIPT | EV_ONESHOT, 0, 0, source);
        
        dispatch_kevent_source_t kevent_source = dispatch_source_get_global_kevent_source();
        kevent(kevent_source->kq, &event, 1, &event, 1, NULL);
      }
      
      dispatch_release(source);
    });
  }
  
};

#pragma mark write source

const struct dispatch_source_type_s _dispatch_source_type_write = {
  
  // I'm kinda channeling the linux source code about now ;)
  .registerBlock = ^(dispatch_source_t source, uintptr_t handle, unsigned long mask){
    struct kevent event;
    EV_SET(&event, handle, EVFILT_WRITE, EV_ADD | EV_RECEIPT | EV_ONESHOT, 0, 0, source);
    
    dispatch_kevent_source_t kevent_source = dispatch_source_get_global_kevent_source();
    int res = kevent(kevent_source->kq, &event, 1, &event, 1, NULL);
    if (res < 0)
    {
      WW_ERROR("adding kevent for read source failed: %d, %s\n", errno, strerror(errno));
    }
  },
  
  .unregisterBlock = ^(dispatch_source_t source){
    struct kevent event;
    EV_SET(&event, source->handle, EVFILT_WRITE, EV_DELETE, 0, 0, NULL);
    
    dispatch_kevent_source_t kevent_source = dispatch_source_get_global_kevent_source();
    int res = kevent(kevent_source->kq, &event, 1, &event, 1, NULL);
    if (res < 0)
    {
      WW_ERROR("unregistering kevent failed: %d, %s\n", errno, strerror(errno));
    }
  },
  
  .dispatchEvent = ^(dispatch_source_t source, struct kevent event){
    dispatch_retain(source);
    source->data = event.data;
    
    _dispatch_source_async(source, ^{
      source->event_handler();
      
      if (!source->cancelled && (source->object.suspend_count == 0))
      {
        // once we've handled, we need to put the event back into the event queue
        struct kevent event;
        EV_SET(&event, source->handle, EVFILT_WRITE, EV_ADD | EV_RECEIPT | EV_ONESHOT, 0, 0, source);
        
        dispatch_kevent_source_t kevent_source = dispatch_source_get_global_kevent_source();
        kevent(kevent_source->kq, &event, 1, &event, 1, NULL);
      }      
      dispatch_release(source);
    });
  }
  
};

#pragma mark timer source

const struct dispatch_source_type_s _dispatch_source_type_timer = {
  .registerBlock = ^(dispatch_source_t source, uintptr_t handle, unsigned long mask){
    dispatch_kevent_source_t kevent_source = dispatch_source_get_global_kevent_source();
    
    // we need to make sure we stay around
    dispatch_retain(source);
    
    // take our source and put it on the timer queue. they get sorted by due time.
    dispatch_list_insert_sorted(kevent_source->timers, source, ^(void *a, void *b){
      dispatch_source_t timer_a = a;
      dispatch_source_t timer_b = b;
      
      if (timer_a->start < timer_b->start)
      {
        return -1;
      }
      else if (timer_a->start > timer_b->start)
      {
        return 1;
      }
      return 0;
    });
    
    // hopefully, the above logic has got us our sources sorted by dispatch time. now put
    // the nearest one in the kqueue as a timer. this should shift the current one if it was
    // before the other.
    dispatch_source_t timer_source = dispatch_list_value_at_index(kevent_source->timers, 0);
    
    // we need a timer event in the future. at least within the leeway.
    uint64_t now = mach_absolute_time();
    if (timer_source->start < now)
    {
      int missed = 0;
      
      // now adjust this so we've got a time in the future to get to
      while (timer_source->start < now)
      {
        timer_source->start += timer_source->interval;
        missed++;
      }
      
      timer_source->data = missed;
    }
    
    struct kevent event;
    uint64_t time = timer_source->start - mach_absolute_time();
    EV_SET(&event, 0, EVFILT_TIMER, EV_ADD | EV_RECEIPT | EV_ONESHOT, NOTE_NSECONDS, (time > 0 ? time : 1), timer_source);
    if (kevent(kevent_source->kq, &event, 1, &event, 1, NULL) == -1)
    {
      WW_ERROR("failed to install kevent timer source, %d %s\n", errno, strerror(errno));
    }
  },
  
  .unregisterBlock = ^(dispatch_source_t source){
    // ok, first is to remove the event from the timers list, then we reset the kevent sources so the next best one fires
    dispatch_kevent_source_t kevent_source = dispatch_source_get_global_kevent_source();
    
    // remove
    dispatch_list_remove_value(kevent_source->timers, source);
    
    // disable kevent timers
    struct kevent event;
    EV_SET(&event, 0, EVFILT_TIMER, EV_DELETE | EV_RECEIPT, 0, 0, NULL);
    if (kevent(kevent_source->kq, &event, 1, &event, 1, NULL) == -1)
    {
      WW_ERROR("failed to remove kevent timer source, %d %s\n", errno, strerror(errno));
    }
    
    // now if we've got a timer next in line, put it in the queue instead
    dispatch_source_t next_source = dispatch_list_value_at_index(kevent_source->timers, 0);
    if (next_source)
    {
      dispatch_retain(next_source);
      
      struct kevent event;
      uint64_t time = next_source->start - mach_absolute_time();
      EV_SET(&event, 0, EVFILT_TIMER, EV_ADD | EV_RECEIPT | EV_ONESHOT, NOTE_NSECONDS, (time > 0 ? time : 1), next_source);
      if (kevent(kevent_source->kq, &event, 1, &event, 1, NULL) == -1)
      {
        WW_ERROR("failed to install next timer source, %d %s\n", errno, strerror(errno));
      }
    }
    
    // kevent is removed, now we can release
    dispatch_release(source);
  },
  
  .dispatchEvent = ^(dispatch_source_t source, struct kevent event){
    // balance the retain in the dispatch
    dispatch_retain(source);
    
    // we fired, so take us out of the timer queue
    source->type->unregisterBlock(source);
    
    source->data = 0;
    source->start += source->interval;
    
    _dispatch_source_async(source, ^{
      source->event_handler();
      
      if (!source->cancelled && (source->object.suspend_count == 0))
      {
        source->type->registerBlock(source, source->handle, source->mask);
      }
      
      dispatch_release(source);
    });
  }
};

#pragma mark add source

const struct dispatch_source_type_s _dispatch_source_type_add = {
  // the register and unregister don't do anything here
  .registerBlock = ^(dispatch_source_t source, uintptr_t handle, unsigned long mask){},
  .unregisterBlock = ^(dispatch_source_t source){},
  
  // dispatchEvent will be called by merge data
  .dispatchEvent = ^(dispatch_source_t source, struct kevent event){
    dispatch_retain(source);
    source->data += event.data;
    
    dispatch_async(source->queue, ^{
      source->event_handler();
      
      // once we've read, we reset the data handler. like EV_CLEAR on kevent
      source->data = 0;
    });
  }
};

#pragma mark or source

const struct dispatch_source_type_s _dispatch_source_type_or = {
  // the register and unregister don't do anything here
  .registerBlock = ^(dispatch_source_t source, uintptr_t handle, unsigned long mask){},
  .unregisterBlock = ^(dispatch_source_t source){},
  
  // dispatchEvent will be called by merge data
  .dispatchEvent = ^(dispatch_source_t source, struct kevent event){
    dispatch_retain(source);
    source->data |= event.data;
    
    dispatch_async(source->queue, ^{
      source->event_handler();
      
      // once we've read, we reset the data handler. like EV_CLEAR on kevent
      source->data = 0;
    });
  }
};

#pragma mark source functions

// --

dispatch_source_t dispatch_source_create(dispatch_source_type_t type, uintptr_t handle, unsigned long mask, dispatch_queue_t queue)
{
  dispatch_source_t source = (dispatch_source_t)malloc(sizeof(struct dispatch_source_s));
  memset(source, '\0', sizeof(struct dispatch_source_s));
 
  source->object.retain_count = 1;
  source->object.suspend_count = 1; // we start suspended
  source->object.dealloc = Block_copy(^{
    dispatch_release(source->queue);
    if (source->event_handler != NULL)
    {
      Block_release(source->event_handler);
    }
  });
  source->object.resume = Block_copy(^{
    source->type->registerBlock(source, handle, mask);
  });
  source->object.suspend = Block_copy(^{
    source->type->unregisterBlock(source);
  });
  
  source->cancelled = 0;
  pthread_cond_init(&source->cancel_cond, NULL);
  
  source->handle = handle;
  source->mask = mask;
  source->type = type;
  source->event_handler = NULL;
  source->context = NULL;
  source->queue = queue;
  dispatch_retain(source->queue);
  
  // we don't register yet. wait till we're resumed.
  
  return source;
}

void dispatch_source_cancel(dispatch_source_t source)
{
  // we still need the block around till our cancelation handler
  // has run
  dispatch_retain(source);
  
  OSAtomicCompareAndSwap32Barrier(0, -1, &source->cancelled);

  // unreg
  source->type->unregisterBlock(source);
  
  WW_DBG("unreg %p\n", source);

  dispatch_async(source->queue, ^{
    
    if (source->running != 0)
    {
      pthread_mutex_t mutex;
      pthread_mutex_init(&mutex, NULL);
      
      pthread_cond_wait(&source->cancel_cond, &mutex);
    }
    
    if (source->cancel_handler)
    {
      dispatch_async(source->queue, source->cancel_handler);
    }
    
    dispatch_release(source);
  });
}

void dispatch_source_set_event_handler(dispatch_source_t source, dispatch_block_t handler)
{
  if (source->event_handler)
  {
    Block_release(source->event_handler);
    source->event_handler = NULL;
  }
  
  if (handler != NULL)
  {
    source->event_handler = Block_copy(handler);
  }
}

void dispatch_source_set_cancel_handler(dispatch_source_t source, dispatch_block_t cancel_handler)
{
  if (source->cancel_handler)
  {
    Block_release(source->cancel_handler);
    source->cancel_handler = NULL;
  }
  
  if (cancel_handler != NULL)
  {
    source->cancel_handler = Block_copy(cancel_handler);
  }
}

void dispatch_source_set_timer(dispatch_source_t source, dispatch_time_t start, uint64_t interval, uint64_t leeway)
{
  // these two funcs might want a lock
  source->start = start;
  source->interval = interval;
  source->leeway = leeway;
}

void dispatch_source_merge_data(dispatch_source_t source, unsigned long value)
{
  // merge is kinda easy. make sure value != 0 and then call dispatch event
  if (((source->type == DISPATCH_SOURCE_TYPE_DATA_ADD) || (source->type == DISPATCH_SOURCE_TYPE_DATA_OR)) &&
      value != 0)
  {
    struct kevent event;
    event.data = value;
    
    source->type->dispatchEvent(source, event);
  }
}

unsigned long dispatch_source_get_data(dispatch_source_t source)
{
  return source->data;
}

uintptr_t dispatch_source_get_handle(dispatch_source_t source)
{
  return source->handle;
}

unsigned long dispatch_source_get_mask(dispatch_source_t source)
{
  return source->mask;
}
